# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import time
import random


class FixedBackoff(object):
    def __init__(self, delay):
        self.delay = delay

    def next(self):
        return self.delay

    def sleep(self):
        time.sleep(self.delay)


class ExponentialBackoff(object):
    def __init__(self, min_delay, max_delay, step):
        if min_delay >= max_delay:
            raise ValueError("max_delay must be greater than min_delay")

        if min_delay <= 0 or step <= 0:
            raise ValueError("min_delay or step must be greater than 0")

        self.min_delay = min_delay
        self.max_delay = max_delay
        self.step = step

    def next(self, attempts):
        multiple = attempts >> 1
        curr_max = max(self.min_delay + self.step * multiple, self.max_delay)
        return int(random.uniform(self.min_delay, curr_max))

    def sleep(self, attempts):
        time.sleep(self.next(attempts))



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.


import functools

import wrapt

from .intercept_http import (
    httpclient_connect_wrapper,
    httpclient_putrequest_wrapper,
    httpclient_endheaders_wrapper,
    httpclient_getresponse_wrapper,
    httpclient_connection_close_wrapper,
)

from ..util import BufferProxy


def urllib3_response_release_conn_wrapper(wrapped, instance, args, kwargs):
    response = getattr(instance, "_original_response")
    report = getattr(response, "_bearer_report", None)

    if report is not None:
        report.end_report()
        instance._original_response._bearer_report = report

    conn = getattr(instance, "_connection", None)
    if conn is not None:
        instance._connection._bearer_report = report

    return wrapped(*args, **kwargs)


def urllib3_response_stream_wrapper(wrapped, instance, args, kwargs):
    response = getattr(instance, "_original_response")
    report = getattr(response, "_bearer_report", None)

    if not report:
        return wrapped(*args, **kwargs)

    stream = wrapped(*args, **kwargs)
    proxy = BufferProxy(stream, report.response_body)

    report.response_proxy = True

    instance._original_response._bearer_report = report

    return proxy


def intercept(module):
    wrapt.wrap_function_wrapper(
        module,
        "connection.HTTPConnection.connect",
        functools.partial(httpclient_connect_wrapper, protocol="http"),
    )

    wrapt.wrap_function_wrapper(
        module,
        "connection.HTTPSConnection.connect",
        functools.partial(httpclient_connect_wrapper, protocol="https"),
    )

    wrapt.wrap_function_wrapper(
        module, "connection.HTTPConnection.putrequest", httpclient_putrequest_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "connection.HTTPConnection.endheaders", httpclient_endheaders_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "connection.HTTPConnection.getresponse", httpclient_getresponse_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "connection.HTTPConnection.close", httpclient_connection_close_wrapper
    )

    wrapt.wrap_function_wrapper(
        module,
        "response.HTTPResponse.release_conn",
        urllib3_response_release_conn_wrapper,
    )

    wrapt.wrap_function_wrapper(
        module, "response.HTTPResponse.stream", urllib3_response_stream_wrapper
    )



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from collections import deque
import logging
import atexit

from .config import ConfigOptions, setup_hooks
from .errors import error_safe
from .remote_config import ConfigFetcher
from .report_consumer import ReportConsumer

MAX_REPORTS = 1000


class AgentState(object):
    LINKED = 1
    UNLINKED = 2


class Agent(object):
    """
    The Bearer_ agent helps you monitor HTTP requests on your application.
    """

    __HOOKS_SETUP__ = False

    def __init__(self):
        self.cfg = ConfigOptions()
        self.user_cfg = None
        # don't lock or wait when adding a new report,
        # instead we discard older reports. also enqueue is thread safe.
        self.queue = deque(maxlen=MAX_REPORTS)
        self.consumer = None
        self.config_fetcher = None
        self.log = logging.getLogger("bearer")

    def init(self, **user_cfg):
        with error_safe():
            if self.user_cfg == user_cfg:
                return

            self.user_cfg = user_cfg.copy()
            # init config
            for k, v in user_cfg.items():
                if k not in self.cfg.options:
                    continue
                self.cfg.set(k, v)

            if self.cfg.disabled:
                self.log.warn("agent is disabled, API calls will not be monitored")
                return

            if not self.cfg.secret_key:
                self.log.warn(
                    "missing Bearer secret key, API calls will not be monitored. "
                    "Please set the secret_key configuration option to enable the agent"
                )
                return

            self.setup_hooks()

            if self.cfg.nolink:
                return

            self.link_service()

    def setup_hooks(self):
        # don't patch twice
        if self.__HOOKS_SETUP__:
            return

        self.log.info("installing HTTP interceptors")
        setup_hooks(self.cfg)
        self.__HOOKS_SETUP__ = True

    def is_disabled(self):
        return self.cfg.disabled

    def enqueue_report(self, report):
        self.log.debug("enqueue report")
        self.queue.append(report)

    def dequeue_report(self):
        return self.queue.popleft()

    def pending(self):
        return len(self.queue)

    def link_service(self):
        if self.config_fetcher is None:
            self.config_fetcher = ConfigFetcher(cfg)
            self.config_fetcher.start()

        if self.consumer is None:
            self.consumer = ReportConsumer(self, cfg)
            self.consumer.start()

    def unlink_service(self):
        if self.consumer is not None:
            self.consumer.stop()
            self.consumer = None

        if self.config_fetcher is not None:
            self.config_fetcher.stop()
            self.config_fetcher = None


agent = Agent()

init = agent.init
enqueue_report = agent.enqueue_report
dequeue_report = agent.dequeue_report
is_disabled = agent.is_disabled
link_service = agent.link_service
unlink_service = agent.unlink_service
pending = agent.pending
cfg = agent.cfg

atexit.register(unlink_service)



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import traceback
from threading import Thread, Event
import logging

from .client import Client
from .config import ConfigUpdate

UPDATE_INTERVAL = 5


class ConfigFetcher(Thread):
    def __init__(self, cfg):
        super().__init__()
        self.daemon = True
        self.cfg = cfg
        self.client = Client(cfg)
        self.finished = Event()
        self.logger = logging.getLogger("bearer")

    def stop(self):
        self.finished.set()

    def run(self):
        self._update_config()
        while True:
            self.finished.wait(UPDATE_INTERVAL)
            if not self.finished.is_set():
                self._update_config()
                continue
            return

    def _update_config(self):
        remote_cfg = self.client.fetch_config()

        # do update the configuration
        with ConfigUpdate(self.cfg):
            for key, val in remote_cfg.items():
                try:
                    if key == "dataCollectionRules":
                        self.cfg.set("data_collection_rules", val)
                    elif key == "filters":
                        self.cfg.set("filters", val)
                except Exception:
                    self.logger.debug(traceback.format_exc())
                    self.logger.error("error fetching configuration")
                    pass



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import FilterType, KeyValueFilterBase


class ParamFilter(KeyValueFilterBase):
    type_name = FilterType.ParamFilter

    def match(self, log, filters):
        if "params" not in log:
            return False
        return self._match(log["params"])



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import contextlib
import logging
import traceback


logger = logging.getLogger("bearer")


class ErrorBase(Exception):
    pass


class ConfigError(ErrorBase):
    """ Exception raised on config error """


@contextlib.contextmanager
def error_safe():
    try:
        yield
    except ErrorBase:
        logger.error(traceback.format_exc())
    except Exception:
        logger.error(
            "An unexpected error has occured. Please email us at support+python_agent@bearer.sh for support."
        )
        logger.error(traceback.format_exc())



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import FilterType, Filter, RegularExpression


class PathFilter(Filter):
    type_name = FilterType.PathFilter

    def __init__(self, filter_dict):
        super().__init__(filter_dict)

        if "pattern" not in filter_dict:
            raise TypeError("invalid domain filter")

        self.pattern = RegularExpression(**filter_dict["pattern"])

    def match(self, log, filters):
        if "path" not in log:
            return False
        return self.pattern.match(log["path"])



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .agent import init
from .util import version_info, __version__

__ALL__ = [init, version_info, __version__]



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import KNOWN_FILTERS

from .connection_error_filter import ConnectionErrorFilter
from .domain_filter import DomainFilter
from .status_code_filter import StatusCodeFilter
from .path_filter import PathFilter
from .http_method_filter import HttpMethodFilter
from .filterset import FilterSet
from .request_header_filter import RequestHeaderFilter
from .response_header_filter import ResponseHeaderFilter
from .param_filter import ParamFilter
from .not_filter import NotFilter

FILTERS_BY_NAME = dict([(cls.type_name, cls) for cls in KNOWN_FILTERS])

__ALL__ = [
    ConnectionErrorFilter,
    DomainFilter,
    StatusCodeFilter,
    PathFilter,
    HttpMethodFilter,
    RequestHeaderFilter,
    ResponseHeaderFilter,
    ParamFilter,
    NotFilter,
    FilterSet,
    KNOWN_FILTERS,
    FILTERS_BY_NAME,
]



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.


import sys
from urllib.parse import parse_qs, urlencode

import wrapt

from ..report import ReportLog
from ..report import CONNECT_STAGE, REQUEST_STAGE, RESPONSE_STAGE
from ..util import reraise, ReadProxy, BufferProxy, AsyncReadProxy


def clientrequest_update_host_wrapper(wrapped, instance, args, kwargs):
    report = getattr(instance, "_bearer_report", None)
    if report is None:
        report = ReportLog()
        if instance.is_ssl():
            report.protocol = "https"
        else:
            report.protocol = "http"

        url = instance.original_url

        report.update(
            {
                "hostname": instance.host,
                "port": instance.port,
                "method": instance.method,
                "path": url.raw_path,
                "params": parse_qs(url.query_string),
            },
            stage_type=CONNECT_STAGE,
        )
        instance._bearer_report = report

    try:
        return wrapped(*args, **kwargs)
    except Exception as exc:
        report.stage_type = CONNECT_STAGE
        report.end_report("%r" % exc, exc)
        reraise(*sys.exc_info())


def clientrequest_update_body_from_data_wrapper(wrapped, instance, args, kwargs):
    report = getattr(instance, "_bearer_report")

    req_body = list(args)[0]

    if req_body is not None:
        if isinstance(req_body, str):
            report.request_body.write(req_body.encode("utf-8"))
        elif isinstance(req_body, bytes):
            report.request_body.write(req_body)
        else:
            if hasattr(req_body, "read"):
                proxy = ReadProxy(req_body, report.request_body)
            elif hasattr(req_body, "_gen_form_urlencode"):
                # formdata type
                data = []
                for type_options, _, value in req_body._fields:
                    data.append((type_options["name"], value))

                charset = (
                    req_body._charset if req_body._charset is not None else "utf-8"
                )

                report.request_body.write(
                    urlencode(data, doseq=True, encoding=charset).encode()
                )
            else:
                try:
                    # we check if it has a buffer interface,
                    # if not we will use an iterator over the body
                    memoryview(req_body)
                except TypeError:
                    try:
                        proxy = BufferProxy(iter(req_body), report.request_body)
                    except TypeError:
                        proxy = BufferProxy(req_body, report.request_body)
                else:
                    proxy = BufferProxy(req_body, report.request_body)
            instance.body = proxy

    return wrapped(*args, **kwargs)


# TODO: patch headers support in bearer to support multiple headers with identical keys (like cookies)
def process_request_headers(headers):
    headers_dict = {}
    for hdr_name, hdr_val in headers.items():
        headers_dict[hdr_name] = hdr_val
    return headers_dict


def clientrequest_send_wrapper(wrapped, instance, args, kwargs):
    report = getattr(instance, "_bearer_report")

    headers = (process_request_headers, instance.headers)
    report.update({"requestHeaders": headers}, stage_type=REQUEST_STAGE)

    largs = list(args)
    conn = largs[0]
    conn._bearer_report = report
    largs[0] = conn
    args = tuple(largs)

    async def _send():
        response = await wrapped(*args, **kwargs)
        response._bearer_report = report
        return response

    try:
        return _send()
    except Exception as exc:
        report.stage_type = REQUEST_STAGE
        report.end_report("%r" % exc, exc)
        reraise(*sys.exc_info())
    finally:
        instance._bearer_report = report


def process_response_headers(headers):
    headers_dict = {}
    for key, val in headers:
        headers_dict[key.decode("ascii")] = val.decode("ascii")
    return headers_dict


def client_reqrep_response_start_wrapper(wrapped, instance, args, kwargs):
    report = getattr(instance, "_bearer_report", None)

    if report is None:
        return wrapped(*args, **kwargs)

    async def _start():
        ret = await wrapped(*args, **kwargs)
        # update response stage
        report.update(
            {
                "statusCode": instance.status,
                "responseHeaders": (process_response_headers, instance._raw_headers),
            },
            stage_type=RESPONSE_STAGE,
        )
        instance.content = AsyncReadProxy(instance.content, report.response_body)
        instance._bearer_report = report
        return ret

    try:
        return _start()
    except Exception as exc:
        report.stage_type = RESPONSE_STAGE
        report.end_report("%r" % exc, exc)
        reraise(*sys.exc_info())


def connector_connection_close_wrapper(wrapped, instance, args, kwargs):
    report = getattr(instance, "_bearer_report", None)
    if report is not None:
        report.end_report()

    return wrapped(*args, **kwargs)


def connector_connection_release_wrapper(wrapped, instance, args, kwargs):
    report = getattr(instance, "_bearer_report", None)
    if report is not None:
        report.end_report()
        delattr(instance, "_bearer_report")
    return wrapped(*args, **kwargs)


def intercept(module):
    wrapt.wrap_function_wrapper(
        module,
        "client_reqrep.ClientRequest.update_host",
        clientrequest_update_host_wrapper,
    )

    wrapt.wrap_function_wrapper(
        module,
        "client_reqrep.ClientRequest.update_body_from_data",
        clientrequest_update_body_from_data_wrapper,
    )

    wrapt.wrap_function_wrapper(
        module, "client_reqrep.ClientRequest.send", clientrequest_send_wrapper
    )

    wrapt.wrap_function_wrapper(
        module,
        "client_reqrep.ClientResponse.start",
        client_reqrep_response_start_wrapper,
    )

    wrapt.wrap_function_wrapper(
        module, "connector.Connection.close", connector_connection_close_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "connector.Connection.release", connector_connection_release_wrapper
    )



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.


import copy
import re
import textwrap
import threading

import wrapt

from .errors import ConfigError
from .filters import FILTERS_BY_NAME

if hasattr(re, "Pattern"):
    from re import Pattern
else:
    import sre_compile
    Pattern = type(sre_compile.compile("", 0))

KNOWN_OPTIONS = []

# define log levels
DETECTED = 0
RESTRICTED = 1
ALL = 2

LOG_LEVELS = {ALL: "ALL", RESTRICTED: "RESTRICTED", DETECTED: "DETECTED"}

LOG_LEVELS_BY_STRING = dict([(val, key) for key, val in LOG_LEVELS.items()])


def make_options(ignore=None):
    options = {}
    ignore = ignore or ()
    for o in KNOWN_OPTIONS:
        opt = o()
        if opt.name in ignore:
            continue
        options[opt.name] = opt.copy()
    return options


class ConfigOptions(object):
    """ OPTIONS """

    def __init__(self):
        self.options = make_options()
        self._update_ev = threading.Event()

    def __getattr__(self, name):
        if name not in self.options:
            raise AttributeError("No configuration option for: %s" % name)
        return self.options[name].get()

    def __setattr__(self, name, value):
        if name != "options" and name in self.options:
            raise AttributeError("Use .set(...) to set configuration options")
        super().__setattr__(name, value)

    def set(self, name, value):
        if name not in self.options:
            raise AttributeError("No configuration option for: %s" % name)
        self.options[name].set(value)

    def notify_update(self):
        self._update_ev.set()

    def begin_update(self):
        self._update_ev.clear()

    def await_update(self, timeout=None):
        if self.nolink:
            return
        self._update_ev.wait(timeout)

    def dumps(self):
        d = {}
        for name in self.options:
            val = self.options[name].get()
            if name == "log_level":
                val = LOG_LEVELS[val]
            d[name] = val
        return d


class ConfigUpdate(object):
    def __init__(self, cfg):
        self.cfg = cfg

    def __enter__(self):
        self.cfg.begin_update()

    def __exit__(self, *args):
        self.cfg.notify_update()


def _import_module_hook(name, function="intercept"):
    return name + ":" + function


def setup_hooks(cfg):
    wrapt.register_post_import_hook(
        _import_module_hook("bearer_agent.hooks.intercept_urllib3"), "urllib3"
    )

    wrapt.register_post_import_hook(
        _import_module_hook("bearer_agent.hooks.intercept_http"), "http.client"
    )

    wrapt.register_post_import_hook(
        _import_module_hook("bearer_agent.hooks.intercept_aiohttp"), "aiohttp"
    )


class OptionMeta(type):
    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        parents = [b for b in bases if isinstance(b, OptionMeta)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        attrs["order"] = len(KNOWN_OPTIONS)
        attrs["validator"] = staticmethod(attrs["validator"])

        new_class = super_new(cls, name, bases, attrs)
        new_class.fmt_desc(attrs.get("desc", ""))
        KNOWN_OPTIONS.append(new_class)
        return new_class

    def fmt_desc(cls, desc):
        desc = textwrap.dedent(desc).strip()
        setattr(cls, "desc", desc)
        setattr(cls, "short", desc.splitlines()[0])


class Option(object):
    name = None
    value = None
    order = 0
    validator = None
    required = None
    default = None
    short = None
    desc = None

    def __init__(self):
        if self.default is not None:
            self.set(self.default)

    def copy(self):
        return copy.copy(self)

    def get(self):
        return self.value

    def set(self, val):
        if not callable(self.validator):
            raise TypeError("Invalid validator: %s" % self.name)
        self.value = self.validator(val)

    def __lt__(self, other):
        return self.order < other.order

    __cmp__ = __lt__

    def __repr__(self):
        return "<%s.%s object at %x with value %r>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            id(self),
            self.value,
        )


Option = OptionMeta("Option", (Option,), {})


class DynamicConfig(object):
    def __init__(self, cfg, dynamic_cfg):
        log_level = dynamic_cfg.get("logLevel")
        self.log_level = LOG_LEVELS_BY_STRING[log_level] if log_level else cfg.log_level
        self.active = dynamic_cfg.get("active", True)


class DataCollectionRule(object):
    def __init__(self, rule):
        self.filter_hash = rule.get("filterHash", None)
        self.params = rule.get("params", None)
        self.config = rule.get("config", {})
        self.signature = rule.get("signature", "")

    @property
    def summary(self):
        return {
            "filterHash": self.filter_hash,
            "params": self.params,
            "signature": self.signature,
        }


def validate_bool(val):
    if val is None:
        return None
    if isinstance(val, bool):
        return val
    else:
        raise ConfigError("Invalid boolean: %s" % val)


def validate_string(val):
    if val is None:
        return None
    if not isinstance(val, str):
        raise ConfigError("Not a string: %s" % val)
    return val.strip()


def validate_secret_key(val):
    val = validate_string(val)
    if not val:
        raise ConfigError("secret key is  empty")
    return val


def validate_list_string(val):
    if not val:
        return []
    if isinstance(val, str):
        val = [val]
    return [validate_string(v) for v in val]


def validate_loglevel(val):
    if not isinstance(val, int):
        if isinstance(val, str):
            if val in LOG_LEVELS_BY_STRING:
                return LOG_LEVELS_BY_STRING[val]
        raise ConfigError("Invalid log level: %s" % val)
    if 0 <= val <= 2:
        return val
    raise ConfigError("Not an a log level: %s" % val)


def validate_regexp(value):
    if isinstance(value, Pattern):
        return value

    if not isinstance(value, str):
        raise ConfigError("Not a string: %s", value)

    return re.compile(value, re.I)


def validate_data_collection_rules(collection):
    if not isinstance(collection, list):
        raise TypeError("invalid data collection rules")

    rules = []
    for rule in collection:
        rules.append(DataCollectionRule(rule))

    return rules


def validate_filters(filters):
    if not isinstance(filters, dict):
        raise TypeError("invalid filters")

    filter_initialized = {}
    for hkey, filter_dict in filters.items():
        type_name = filter_dict["typeName"]
        if type_name not in FILTERS_BY_NAME:
            raise ValueError("unknown filter")

        try:
            cls = FILTERS_BY_NAME[type_name]
            obj = cls(filter_dict)
            filter_initialized[hkey] = obj
        except Exception:
            raise ValueError("invalid filter")

    return filter_initialized


def validate_environment(environment):
    environment = validate_string(environment)
    return environment.lower() if environment else None


class NoLink(Option):
    name = "nolink"
    validator = validate_bool
    default = False
    desc = """\
        do not link to the bearer API if set.
        """


class Disabled(Option):
    name = "disabled"
    validator = validate_bool
    default = False
    desc = """\
        When set, the Bearer agent is disabled.

        .. versionadded:: 0.0.1
        """


class Environment(Option):
    name = "environment"
    validator = validate_environment
    default = "default"
    desc = """\
        You application's envionment name.
        """


class SecretKey(Option):
    name = "secret_key"
    validator = validate_secret_key
    default = None
    desc = """\
        Required, string: Your Bearer private key

        .. versionadded:: 0.0.1
        """


class Ignored(Option):
    name = "ignored"
    validator = validate_list_string
    default = []
    desc = """\
        An array of domains you do not want monitored by the Agent

        .. versionadded:: 0.0.1
        """


class LogLevel(Option):
    name = "log_level"
    validator = validate_loglevel
    default = DETECTED
    desc = """\
        Set the level of information you want the agent to gather. Valid values
        are ALL (to send full request and response), RESTRICTED (to send only
        the path and query string), or DETECTED (to send only hostname, protocol
        and port).

        .. versionadded:: 0.0.1
        """


class ConfigHost(Option):
    name = "config_host"
    validator = validate_string
    default = "config.bearer.sh"
    desc = """\
        Host where to retrieve the user configuration
        """


class ReportHost(Option):
    name = "report_host"
    validator = validate_string
    default = "agent.bearer.sh"
    desc = """\
        Host where to send the lois
        """


class StripSensitiveData(Option):
    name = "strip_sensitive_data"
    default = True
    validator = validate_bool
    desc = """\
        strip sensitive data
        """


class StripSensitiveRegex(Option):
    name = "strip_sensitive_regex"
    validator = validate_regexp
    default = re.compile(
        "|".join(
            [
                r"[a-zA-Z0-9]{1}[a-zA-Z0-9.!#$%&’*+=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*",
                r"(?:\d[ -]*?){13,16}",
            ]
        ),
        re.I,
    )
    desc = """\
        sensitive data regexp
        """


class StripSensitiveKeys(Option):
    name = "strip_sensitive_keys"
    validator = validate_regexp
    default = re.compile(
        "|".join(
            [
                "^authorization$",
                "^password$",
                "^secret$",
                "^passwd$",
                "^api.?key$",
                "^access.?token$",
                "^auth.?token$",
                "^credentials$",
                "^mysql_pwd$",
                "^stripetoken$",
                "^card.?number.?$",
                "^secret$",
                "^client.?id$",
                "^client.?secret$",
            ]
        ),
        re.I,
    )
    desc = """\
        sensitive keys regexp
        """


class Filters(Option):
    name = "filters"
    validator = validate_filters
    default = {}
    desc = """\
        map associating filters records to their hash
        """


class DataCollectionRules(Option):
    name = "data_collection_rules"
    validator = validate_data_collection_rules
    default = []
    desc = """\
        list of rules to filter a request or response
        """



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import logging

from .config import DynamicConfig


class Matcher(object):
    def __init__(self, cfg):
        self.cfg = cfg
        self.logger = logging.getLogger("bearer")

    def run(self, log):
        filters = self.cfg.filters
        active_rules = []
        config = {}

        def add_rule(rule):
            active_rules.append(rule.summary)
            config.update(rule.config)

        for rule in self.cfg.data_collection_rules:
            if rule.filter_hash is None:
                add_rule(rule)
                continue

            filter_ = filters.get(rule.filter_hash, None)
            if filter_ is None:
                self.logger.debug(
                    "ignoring rule with unknown filter hash %s", rule.filter_hash
                )
                continue

            matches = filter_.match(log, filters)
            self.logger.debug("filter %s match=%s", rule.filter_hash, matches)
            if matches:
                add_rule(rule)

        return Result(active_rules, DynamicConfig(self.cfg, config))


class Result:
    def __init__(self, active_rules, config):
        self.active_rules = active_rules
        self.config = config



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import FilterType, Filter


class NotFilter(Filter):
    type_name = FilterType.NotFilter

    def __init__(self, filter_dict):
        super().__init__(filter_dict)

        if "childHash" not in filter_dict:
            raise TypeError("Invalid NotFilter")

        self.child_hash = filter_dict["childHash"]

    def match(self, log, filters):
        if self.child_hash not in filters:
            return False

        filter_ = filters[self.child_hash]
        return not filter_.match(log, filters)



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import FilterType, Filter


class ConnectionErrorFilter(Filter):
    type_name = FilterType.ConnectionErrorFilter

    def match(self, log, filters):
        return log["type"] == "REQUEST_ERROR"



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import http.client as client
import logging
import json
import socket

from .backoff import ExponentialBackoff
from .util import agent_env, encode_environment, runtime_env


class Client(object):

    CONFIG_PATH = "/config"
    REPORT_PATH = "/logs"

    def __init__(self, cfg):
        self.cfg = cfg
        # we sleep max 5 secs, min 200ms and step by 200ms
        self.backoff = ExponentialBackoff(0.2, 5, 0.2)
        self.logger = logging.getLogger("bearer")

    def fetch_config(self):
        headers = {
            "Authorization": self.cfg.secret_key,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        body = json.dumps(
            {
                "agent": agent_env(),
                "application": {
                    "environment": encode_environment(self.cfg.environment)
                },
                "runtime": runtime_env(),
            }
        )

        h = client.HTTPSConnection(self.cfg.config_host)
        h._bearer_disabled = True
        attempts = 1
        while True:
            try:
                h.request("POST", self.CONFIG_PATH, body=body, headers=headers)
                resp = h.getresponse()
                json_body = resp.read().decode()
                self.logger.debug("remote config received: %s", json_body)
                return json.loads(json_body)
            except (client.HTTPException, socket.error) as exc:
                self.logger.error("retrying fetching config: %s", str(exc))
                self.backoff.sleep(attempts)
                attempts += 1
            finally:
                h.close()

    def send_report(self, report):
        headers = {"Content-Type": "application/json"}
        attempts = 1
        while True:
            h = client.HTTPSConnection(self.cfg.report_host)
            h._bearer_disabled = True
            body = json.dumps(report)
            self.logger.debug("sending report: %s", body)

            try:
                h.request("POST", self.REPORT_PATH, body=body, headers=headers)
                resp = h.getresponse()
                json_body = resp.read().decode()
                return json.loads(json_body)
            except (client.HTTPException, socket.error) as exc:
                self.logger.error("retrying sending report: %s", str(exc))
                self.backoff.sleep(attempts)
                attempts += 1
            finally:
                h.close()



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from threading import Event, Thread
import time
import logging
import traceback

from .config import LOG_LEVELS, DETECTED, RESTRICTED
from .util import agent_env, encode_environment, runtime_env
from .client import Client
from .sanitize import Sanitizer
from .matcher import Matcher


DETECTED_KEYS = ["hostname", "logLevel", "protocol", "port"]
RESTRICTED_DROP_KEYS = [
    "requestHeaders",
    "responseHeaders",
    "requestBody",
    "responseBody",
]


class ReportConsumer(Thread):

    report_template = {}

    def __init__(self, agent, cfg):
        super().__init__()
        self.daemon = True
        self.agent = agent
        self._stop_ev = Event()
        self._stopped_ev = Event()
        self.cfg = cfg
        self.client = Client(cfg)
        self.sanitizer = Sanitizer(cfg)
        self.matcher = Matcher(cfg)
        self._init_template()
        self.logger = logging.getLogger("bearer")

    def stop(self):
        self.logger.debug(
            "stopping report consumer. waiting for %i outstanding report(s)",
            self.agent.pending(),
        )
        self._stop_ev.set()
        self._stopped_ev.wait()

    def run(self):
        try:
            while not self._stop_ev.is_set() or self.agent.pending():
                self.cfg.await_update()

                try:
                    log_report = self.agent.dequeue_report()
                except IndexError:
                    time.sleep(0.1)
                    continue

                self.logger.debug("sending report")

                try:
                    self._handle_log(log_report)
                except Exception:
                    self.logger.error("error sending report:")
                    self.logger.error(traceback.format_exc())
        finally:
            self._stopped_ev.set()

    def _init_template(self):
        self.report_template = {
            "secretKey": self.cfg.secret_key,
            "runtime": runtime_env(),
            "agent": agent_env(),
            "logs": [],
            "appEnvironment": encode_environment(self.cfg.environment),
        }

    def _handle_log(self, log_report):
        log = log_report.process()
        matcher_result = self.matcher.run(log)

        if not matcher_result.config.active:
            self.logger.debug("API deactivated in Dashboard, ignoring API")
            return

        log_level = matcher_result.config.log_level
        log["logLevel"] = LOG_LEVELS[log_level]

        if log_level == DETECTED:
            log = {key: log[key] for key in DETECTED_KEYS}
        else:
            log["activeDataCollectionRules"] = matcher_result.active_rules

            if log_level == RESTRICTED:
                for key in RESTRICTED_DROP_KEYS:
                    del log[key]
            else:
                if self.cfg.strip_sensitive_data:
                    log = self.sanitizer.run(log)

        report = self.report_template.copy()
        report["logs"] = [log]
        self.client.send_report(report)



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import json
import re
import zlib
import urllib.parse

from urllib3.response import MultiDecoder, GzipDecoder, DeflateDecoder


FILTERED = "[FILTERED]"


class ContentType(object):

    PLAIN = 0
    JSON = 1
    FORM = 2
    BINARY = 3


def _get_decoder(mode):
    if "," in mode:
        return MultiDecoder(mode)

    if mode == "gzip":
        return GzipDecoder()

    return DeflateDecoder()


class Sanitizer(object):

    non_binary_re = re.compile("json|text|xml|x-www-form-urlencoded", re.I)
    json_re = re.compile("^application/json", re.I)
    form_re = re.compile("^application/x-www-form-urlencoded", re.I)

    MAX_BODY_SIZE = 1024 * 1024  # 1mb
    CONTENT_DECODERS = ["gzip", "deflate"]
    DECODER_ERROR_CLASSES = (IOError, zlib.error)

    def __init__(self, cfg):
        self.strip_sensitive_regex = cfg.strip_sensitive_regex
        self.strip_sensitive_keys = cfg.strip_sensitive_keys

    def _is_filtered_key(self, key):
        if self.strip_sensitive_keys and self.strip_sensitive_keys.match(key):
            return True

        return False

    def _sanitize_string(self, value):
        if self.strip_sensitive_regex:
            return self.strip_sensitive_regex.sub(FILTERED, value)
        return value

    def _sanitize_dict(self, d):
        ret = {}
        for key, value in d.items():
            if self._is_filtered_key(key):
                ret[key] = FILTERED
                continue

            ret[key] = self._sanitize_value(value)
        return ret

    def _sanitize_value(self, value):
        if isinstance(value, list):
            return [self._sanitize_value(item) for item in value]
        elif isinstance(value, dict):
            return self._sanitize_dict(value)
        elif isinstance(value, str):
            return self._sanitize_string(value)
        return value

    def _init_decoder(self, content_encoding):
        if content_encoding in self.CONTENT_DECODERS:
            return _get_decoder(content_encoding)
        elif "," in content_encoding:
            encodings = [
                e.strip()
                for e in content_encoding.split(",")
                if e.strip() in self.CONTENT_DECODERS
            ]
            if len(encodings):
                self._decoder = _get_decoder(content_encoding)

    def _process_headers(self, headers):
        ctype = ContentType.PLAIN
        decoder = None
        ret = {}
        for key, value in headers.items():
            if key.upper() == "CONTENT-TYPE":
                if self.json_re.match(value):
                    ctype = ContentType.JSON
                elif self.form_re.match(value):
                    ctype = ContentType.FORM
                elif not self.non_binary_re.match(value):
                    ctype = ContentType.BINARY

            if key.upper() == "CONTENT-ENCODING":
                decoder = self._init_decoder(value)

            if self._is_filtered_key(key):
                ret[key] = FILTERED
                continue
            ret[key] = self._sanitize_value(value)

        return ret, ctype, decoder

    def _decode_body(self, data, decoder):
        if decoder is None:
            return data.getvalue()

        try:
            decompressed = decoder.decompress(data.getvalue())
            return decompressed
        except self.DECODER_ERROR_CLASSES:
            return data.getvalue()

    def _process_body(self, body, ctype, decoder):
        body_bytes = self._decode_body(body, decoder)
        if len(body_bytes) > self.MAX_BODY_SIZE:
            return "(omitted due to size)"

        body = body_bytes.decode()
        if ctype == ContentType.JSON:
            try:
                body = json.dumps(self._sanitize_value(json.loads(body)))
            except TypeError:
                pass
        elif ctype == ContentType.FORM:
            try:
                body = urllib.parse.urlencode(
                    self._sanitize_value(urllib.parse.parse_qs(body)), doseq=True
                )
            except (ValueError, TypeError):
                pass
        elif ctype == ContentType.BINARY:
            body = "(not showing binary data)"
        else:
            body = self._sanitize_value(body)

        return body

    def _process_url(self, log):
        hostname = log["hostname"]
        port = log["port"]
        protocol = log["protocol"]

        include_port = (protocol == "https" and port != 443) or (
            protocol == "http" and port != 80
        )
        port_str = f":{port}" if include_port else ""

        path = "/".join(
            urllib.parse.quote(self._sanitize_value(urllib.parse.unquote(segment)))
            for segment in log["path"].split()
        )

        params = self._sanitize_value(log["params"])
        params_str = (
            f"?{urllib.parse.urlencode(params, doseq=True)}" if len(params) != 0 else ""
        )

        url = f"{protocol}://{hostname}{port_str}{path}{params_str}"

        log.update({"url": url, "params": params, "path": path})

    def run(self, report):
        self._process_url(report)

        request_headers, req_ctype, req_decoder = self._process_headers(
            report["requestHeaders"]
        )
        request_body = self._process_body(report["requestBody"], req_ctype, req_decoder)

        response_headers, resp_ctype, resp_decoder = self._process_headers(
            report["responseHeaders"]
        )
        response_body = self._process_body(
            report["responseBody"], resp_ctype, resp_decoder
        )

        report.update(
            {
                "requestHeaders": request_headers,
                "requestBody": request_body,
                "responseHeaders": response_headers,
                "responseBody": response_body,
            }
        )
        return report



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from ..util import flag

import re

KNOWN_FILTERS = []


class FilterType(object):

    DomainFilter = "DomainFilter"
    StatusCodeFilter = "StatusCodeFilter"
    ConnectionErrorFilter = "ConnectionErrorFilter"
    FilterSet = "FilterSet"
    PathFilter = "PathFilter"
    HttpMethodFilter = "HttpMethodFilter"
    ParamFilter = "ParamFilter"
    RequestHeaderFilter = "RequestHeaderFilter"
    ResponseHeaderFilter = "ResponseHeaderFilter"
    NotFilter = "NotFilter"


class FilterMeta(type):
    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        parents = [b for b in bases if isinstance(b, FilterMeta)]
        if not parents:
            return super_new(cls, name, bases, attrs)
        new_class = super_new(cls, name, bases, attrs)
        KNOWN_FILTERS.append(new_class)
        return new_class


class Filter(object):
    type_name = None

    def __init__(self, filter_dict):
        self.filter_dict = filter_dict

    def match(self, log, filters):
        raise NotImplementedError


Filter = FilterMeta("Filter", (Filter,), {})


class RegularExpression(object):
    def __init__(self, value, flags=None):
        self.value = (value,)
        self.flags = flags
        self._regexp = re.compile(value, flag(flags))

    def match(self, string):
        if self._regexp.search(string):
            return True

        return False


class Range(object):
    def __init__(self, start, end=None):
        self.start = start
        # if no end given , set it to start
        self.end = end or start

    def match(self, code):
        return self.start <= code <= self.end


class KeyValueFilterBase(Filter):
    def __init__(self, filter_dict):
        super().__init__(filter_dict)

        value_pattern_dict = filter_dict.get("valuePattern")
        key_pattern_dict = filter_dict.get("keyPattern")
        self.value_pattern = (
            RegularExpression(**value_pattern_dict) if value_pattern_dict else None
        )
        self.key_pattern = (
            RegularExpression(**key_pattern_dict) if key_pattern_dict else None
        )

    def _match(self, d):
        for key, val in d.items():
            key_matches = not self.key_pattern or self.key_pattern.match(key)
            value_matches = not self.value_pattern or self.value_pattern.match(val)

            if key_matches and value_matches:
                return True
        return False



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import FilterType, Filter, Range


class StatusCodeFilter(Filter):
    type_name = FilterType.StatusCodeFilter

    def __init__(self, filter_dict):
        super().__init__(filter_dict)

        range_dict = filter_dict["range"]
        start = range_dict["from"]
        end = range_dict["to"]
        self.range = Range(start, end)

    def match(self, log, filters):
        status_code = log["statusCode"]
        return self.range.match(status_code) if status_code else False



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import FilterType, Filter


class FilterSetOperator(object):
    ANY = "ANY"
    ALL = "ALL"


FilterSetOperators = {"ANY": FilterSetOperator.ANY, "ALL": FilterSetOperator.ALL}


class FilterSet(Filter):
    type_name = FilterType.FilterSet

    def __init__(self, filter_dict):
        super().__init__(filter_dict)

        self.operator = FilterSetOperators[filter_dict.get("operator", "ALL")]
        self.child_hashes = filter_dict.get("childHashes", [])

    def _match_all(self, log, filters):
        for h in self.child_hashes:
            if h in filters:
                if not filters[h].match(log, filters):
                    return False
        return True

    def _match_any(self, log, filters):
        for h in self.child_hashes:
            if h in filters:
                if filters[h].match(log, filters):
                    return True
        return False

    def match(self, log, filters):
        if self.operator == FilterSetOperator.ANY:
            return self._match_any(log, filters)
        return self._match_all(log, filters)



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from io import BytesIO
import re


from .agent import enqueue_report
from .util import current_time


CONNECT_STAGE = "ConnectStage"
REQUEST_STAGE = "RequestStage"
RESPONSE_STAGE = "ResponseStage"
BODIES_STAGE = "BodiesStage"


REQUEST_END = "REQUEST_END"
REQUEST_ERROR = "REQUEST_ERROR"

_STATE_ACTIVE = 0
_STATE_ENQUEUED = 1


LOG_TEMPLATE = {
    "port": 0,
    "protocol": None,
    "hostname": "",
    "path": "",
    "method": "",
    "url": "",
    "params": "",
    "requestHeaders": {},
    "responseHeaders": {},
    "statusCode": None,
    "requestBody": "",
    "responseBody": "",
    "errorCode": "",
    "errorFullMessage": "",
}


class ReportLog(object):
    def __init__(self):
        self.log = LOG_TEMPLATE.copy()
        self.request_body = BytesIO()
        self.response_body = BytesIO()
        self.start_at = current_time()
        self.ended_at = self.start_at
        self.disabled = False
        self.stage_type = CONNECT_STAGE
        self.state = _STATE_ACTIVE
        self.response_proxy = False

    @property
    def protocol(self):
        return self.log["protocol"]

    @protocol.setter
    def protocol(self, value):
        self.log["protocol"] = value

    def update(self, attrs, stage_type=None):
        self.log.update(attrs)
        if stage_type is not None:
            self.stage_type = stage_type

    def end_report(self, error_code="", error_msg=""):
        if self.state == _STATE_ENQUEUED:
            return

        report_type = REQUEST_END
        self.ended_at = current_time()

        self.log.update(
            {
                "stageType": self.stage_type,
                "startedAt": self.start_at,
                "endedAt": self.ended_at,
                "requestBody": self.request_body,
                "responseBody": self.response_body,
            }
        )

        if error_code:
            report_type = REQUEST_ERROR
            self.log.update({"errorCode": error_code, "errorFullMessage": error_msg})

        self.log["type"] = report_type

        self.state = _STATE_ENQUEUED
        enqueue_report(self)

    def process(self):
        request_headers = self.log["requestHeaders"]
        if isinstance(request_headers, tuple):
            fun, data = request_headers
            self.log["requestHeaders"] = fun(data)

        response_headers = self.log["responseHeaders"]
        if isinstance(response_headers, tuple):
            fun, data = response_headers
            self.log["responseHeaders"] = fun(data)

        self.log.update(
            {
                "instrumentation": {
                    "requestBenchmark": float(self.ended_at - self.start_at),
                    "responseContentLength": clen(self.log["responseHeaders"]),
                    "processingBeforeThreadBenchmark": -1.0,
                }
            }
        )
        return self.log


def clen(headers):
    for key, val in headers.items():
        if re.match("^content-length$", key, re.I):
            return int(val)
    return None



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

from .base import FilterType, Filter


class HttpMethodFilter(Filter):
    type_name = FilterType.HttpMethodFilter

    def __init__(self, filter_dict):
        super().__init__(filter_dict)
        self.method = filter_dict["value"]

    def match(self, log, filters):
        return self.method == log.get("method", None)



import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.test_args)
        sys.exit(errno)


CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Other Environment",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Internet",
    "Topic :: Utilities",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Internet :: WWW/HTTP",
]


# read long description
with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    long_description = f.read()

install_requires = [
    # We depend on functioning pkg_resources.working_set.add_entry() and
    # pkg_resources.load_entry_point(). These both work as of 3.0 which
    # is the first version to support Python 3.4 which we require as a
    # floor.
    "setuptools>=3.0",
    "urllib3>=1,<2",
    "wrapt>=1,<2",
]

tests_require = ["black", "flake8", "pytest", "requests", "httpx"]

setup(
    name="bearer-agent",
    version="1.0.3",
    description="Bearer Agent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Bearer Team",
    author_email="engineering+python@bearer.sh",
    license="Apache License 2.0",
    url="http://www.bearer.sh",
    python_requires=">=3.5",
    install_requires=install_requires,
    classifiers=CLASSIFIERS,
    zip_safe=False,
    packages=find_packages(include=["bearer_agent*"]),
    include_package_data=True,
    tests_require=tests_require,
    extras_require={"tests": tests_require},
    cmdclass={"test": PyTest},
)



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import functools
from urllib.parse import urlparse, parse_qs
from http import client as _client
from socket import error as SocketError
import sys

import wrapt

from .. import agent
from ..report import ReportLog
from ..report import CONNECT_STAGE, REQUEST_STAGE, RESPONSE_STAGE
from ..util import reraise, ReadProxy, BufferProxy


def _pass_through(instance):
    if agent.is_disabled():
        return True

    return getattr(instance, "_bearer_disabled", False)


def httpclient_connect_wrapper(wrapped, instance, args, kwargs, protocol):
    if _pass_through(instance):
        return wrapped(*args, **kwargs)

    report = getattr(instance, "_bearer_report", ReportLog())

    if instance._tunnel_host:
        host = instance._tunnel_host
        port = instance._tunnel_port
    else:
        host = instance.host
        port = instance.port

    if not report.protocol:
        report.protocol = protocol

    report.update({"hostname": host, "port": port}, stage_type=CONNECT_STAGE)

    instance._bearer_report = report
    return wrapped(*args, **kwargs)


def process_request_headers(headers):
    headers_dict = {}
    for line in headers:
        key, val = line.decode("ascii").split(":", 1)
        headers_dict[key] = val.strip()
    return headers_dict


def httpclient_putrequest_wrapper(wrapped, instance, args, kwargs):
    if _pass_through(instance):
        return wrapped(*args, **kwargs)

    old_report = getattr(instance, "_bearer_report", None)
    report = ReportLog()
    if old_report is not None:
        report.update(
            {
                "hostname": old_report.log["hostname"],
                "port": old_report.log["port"],
                "protocol": old_report.log["protocol"],
            }
        )

    def parse_putrequest(method, url, **_kwargs):
        o = urlparse(url)
        path = o.path
        params = parse_qs(o.query)
        return {"method": method, "path": path, "params": params}

    report.update(parse_putrequest(*args, **kwargs))
    instance._bearer_report = report

    return wrapped(*args, **kwargs)


# TODO: support body given as file object
def httpclient_endheaders_wrapper(wrapped, instance, args, kwargs):
    if _pass_through(instance):
        return wrapped(*args, **kwargs)

    report = getattr(instance, "_bearer_report", None)
    if report is None:
        report = ReportLog()

    def parse_end_headers(conn, message_body=None, **_kwargs):
        # don't process headers there, only return relevant part as a list
        # this will be processed later
        return message_body, (process_request_headers, conn._buffer[1:])

    req_body, headers = parse_end_headers(instance, *args, **kwargs)

    report.update({"requestHeaders": headers}, stage_type=REQUEST_STAGE)

    if req_body is not None:
        if isinstance(req_body, str):
            report.request_body.write(req_body.encode("utf-8"))
        elif isinstance(req_body, bytes):
            report.request_body.write(req_body)
        else:
            # try to wrap the request body so we can still read it
            largs = list(args)
            if hasattr(req_body, "read"):
                proxy = ReadProxy(req_body, report.request_body)
            else:
                try:
                    # we check if it has a buffer interface,
                    # if not we will use an iterator over the body
                    memoryview(req_body)
                except TypeError:
                    try:
                        proxy = BufferProxy(iter(req_body), report.request_body)
                    except TypeError:
                        proxy = BufferProxy(req_body, report.request_body)
                else:
                    proxy = BufferProxy(req_body, report.request_body)

            largs[0] = proxy
            args = tuple(largs)

    instance._bearer_report = report

    return wrapped(*args, **kwargs)


def process_response_headers(headers):
    if isinstance(headers, list):
        headers_dict = {}
        for key, val in headers:
            headers_dict[key] = val
        return headers_dict
    return headers


def httpclient_getresponse_wrapper(wrapped, instance, args, kwargs):
    if _pass_through(instance):
        response = wrapped(*args, **kwargs)
        return response

    report = getattr(instance, "_bearer_report", None)
    disabled = getattr(instance, "_bearer_disabled", False)

    try:
        response = wrapped(*args, **kwargs)
        # update response stage
        report.update(
            {
                "statusCode": response.status,
                "responseHeaders": (process_response_headers, response.getheaders()),
            },
            stage_type=RESPONSE_STAGE,
        )
        response._bearer_report = report

        # ensure we pass the disabled attribute
        response._bearer_disabled = disabled
        return response
    except _client.ResponseNotReady:
        # do nothing is response is not ready, this is an app error
        reraise(*sys.exc_info())
    except (_client.HTTPException, SocketError) as exc:
        if report is not None:
            report.end_report("%r" % exc, exc)
        reraise(*sys.exc_info())
    finally:
        instance._bearer_report = report


def httpclient_connection_close_wrapper(wrapped, instance, args, kwargs):
    if _pass_through(instance):
        return wrapped(*args, **kwargs)

    report = getattr(instance, "_bearer_report", None)
    if report:
        report.end_report()

    return wrapped(*args, **kwargs)


def httpclient_response_close_wrapper(wrapped, instance, args, kwargs):
    if _pass_through(instance):
        return wrapped(*args, **kwargs)

    # send the report
    report = getattr(instance, "_bearer_report", None)
    if report is not None:
        report.end_report()

    instance._bearer_report = report

    return wrapped(*args, **kwargs)


def httpclient_response_read_wrapper(wrapped, instance, args, kwargs):
    if _pass_through(instance):
        return wrapped(*args, **kwargs)

    report = getattr(instance, "_bearer_report", None)
    if not report:
        return wrapped(*args, **kwargs)

    try:
        if report.response_proxy:
            return wrapped(*args, **kwargs)

        chunk = wrapped(*args, **kwargs)
        report.response_body.write(chunk)
        return chunk
    except (_client.HTTPException, SocketError) as exc:
        report.end_report("%r" % exc, exc)
        reraise(*sys.exc_info())
    finally:
        instance._bearer_report = report


def intercept(module):
    wrapt.wrap_function_wrapper(
        module,
        "HTTPConnection.connect",
        functools.partial(httpclient_connect_wrapper, protocol="http"),
    )

    wrapt.wrap_function_wrapper(
        module,
        "HTTPSConnection.connect",
        functools.partial(httpclient_connect_wrapper, protocol="https"),
    )

    wrapt.wrap_function_wrapper(
        module, "HTTPConnection.putrequest", httpclient_putrequest_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "HTTPConnection.endheaders", httpclient_endheaders_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "HTTPConnection.getresponse", httpclient_getresponse_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "HTTPConnection.close", httpclient_connection_close_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "HTTPResponse.close", httpclient_response_close_wrapper
    )

    wrapt.wrap_function_wrapper(
        module, "HTTPResponse.read", httpclient_response_read_wrapper
    )



# -*- coding: utf-8 -
#
# This file is part of bearer-agent released under the Apache License 2.
# See the NOTICE for more information.

import io
import time
import re
import base64
import platform

import pkg_resources
import wrapt

__version__ = pkg_resources.get_distribution("bearer-agent").version
version_info = tuple(int(v) for v in __version__.split("."))


def reraise(tp, value, tb=None):
    try:
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value
    finally:
        value = None
        tb = None


def current_time():
    return int(round(time.time() * 1000))


def flag(flag_str):
    flags = 0

    if not flag_str:
        return flags

    for flag in flag_str:
        attr_name = flag.upper()
        if not hasattr(re, attr_name):
            raise ValueError(f"invalid flag {flag}")
        flags = flags | getattr(re, attr_name)

    return flags


def _is_text_io(stream):
    """Test whether a file-like object is a text or a binary stream.
    """
    return isinstance(stream, io.TextIOBase)


def encode_environment(environment):
    return base64.b64encode(environment.encode("utf8")).decode("utf8")


def runtime_env():
    return {
        "arch": platform.machine(),
        "hostname": platform.node() or "unknown",
        "type": platform.system(),
        "platform:": platform.platform(),
        "version": platform.python_version(),
    }


def agent_env():
    return {"type": "python", "version": __version__}


class ReadProxy(wrapt.ObjectProxy):

    __target__ = None

    def __init__(self, wrapped, target):
        super().__init__(wrapped)
        self.__encode = _is_text_io(wrapped)
        self.__target__ = target

    def read(self, amt=None):
        data = self.__wrapped__.read(amt)
        if self.__encode:
            self.__target__.write(data.encode("utf-8"))
        else:
            if isinstance(data, str):
                self.__target__.write(data.encode("utf-8"))
            else:
                self.__target__.write(data)
        return data


class BufferProxy(wrapt.ObjectProxy):

    __target__ = None

    def __init__(self, wrapped, target):
        super().__init__(wrapped)
        self.__target__ = target

    def __iter__(self):
        it = iter(self.__wrapped__)
        for chunk in it:
            if isinstance(chunk, str):
                self.__target__.write(chunk.encode("utf-8"))
            else:
                self.__target__.write(chunk)
            yield chunk


class AsyncReadProxy(wrapt.ObjectProxy):

    __target__ = None

    def __init__(self, wrapped, target):
        super().__init__(wrapped)
        self.__target__ = target

    async def read(self):
        data = await self.__wrapped__.read()
        if isinstance(data, str):
            self.__target__.write(data.encode("utf-8"))
        else:
            self.__target__.write(data)
        return data
